
/*
 * Procedures for handling ispLSI devices.
 * At present we only know about the 1016.
 */
#include <stdio.h>
#include <time.h>
#include "logsyn.h"
#include "pal.h"

extern struct pal *cur_pal;
extern FILE *pal_out;
extern char pal_fname[];
extern char pal_ident[];
extern int verbose;
extern int pdebug;
extern int n_errs;
extern int jedec_cr;

#define MAX_BITMAP_WIDTH	8
typedef unsigned long		bitmap[MAX_BITMAP_WIDTH];

void isp_init(), isp_pin(), isp_spew();

#define FAST_INPUT_PIN	(INPUT_PIN | DED_INPUT)

struct isppininfo {
    int		bits;
    short	bank;
    short	box;
    short	group;
    char	*ioc_name;
} isp1016_pins[] = {
    {GND, 0, 0, -1, "GND"},		{FAST_INPUT_PIN, 1, 0, -1, "I3"},
    {IO_PIN, 1, 4, 8, "IO24"},		{IO_PIN, 1, 4, 9, "IO25"},
    {IO_PIN, 1, 5, 10, "IO26"},		{IO_PIN, 1, 5, 11, "IO27"},
    {IO_PIN, 1, 6, 12, "IO28"},		{IO_PIN, 1, 6, 13, "IO29"},
    {IO_PIN, 1, 7, 14, "IO30"},		{IO_PIN, 1, 7, 15, "IO31"},
    {GLOBAL_CLK, 0, 0, -1, "Y0"},	{VCC, 0, 0, -1, "VCC"},
    {0, 0, 0, -1, "ispEN"},		{FAST_INPUT_PIN, 0, 0, -1, "I0"},
    {IO_PIN, 0, 0, 0, "IO0"},		{IO_PIN, 0, 0, 1, "IO1"},
    {IO_PIN, 0, 1, 2, "IO2"},		{IO_PIN, 0, 1, 3, "IO3"},
    {IO_PIN, 0, 2, 4, "IO4"},		{IO_PIN, 0, 2, 5, "IO5"},
    {IO_PIN, 0, 3, 6, "IO6"},		{IO_PIN, 0, 3, 7, "IO7"},
    {GND, 0, 0, -1, "GND"},		{FAST_INPUT_PIN, 0, 0, -1, "I1"},
    {IO_PIN, 0, 4, 8, "IO8"},		{IO_PIN, 0, 4, 9, "IO9"},
    {IO_PIN, 0, 5, 10, "IO10"},		{IO_PIN, 0, 5, 11, "IO11"},
    {IO_PIN, 0, 6, 12, "IO12"},		{IO_PIN, 0, 6, 13, "IO13"},
    {IO_PIN, 0, 7, 14, "IO14"},		{IO_PIN, 0, 7, 15, "IO15"},
    {GLOBAL_CLK, 0, 0, -1, "Y2"},	{VCC, 0, 0, -1, "VCC"},
    {GLOBAL_CLK|GLOBAL_RESET, 0, 0, -1, "Y1"},
					{FAST_INPUT_PIN, 1, 0, -1, "I2"},
    {IO_PIN, 1, 0, 0, "IO16"},		{IO_PIN, 1, 0, 1, "IO17"},
    {IO_PIN, 1, 1, 2, "IO18"},		{IO_PIN, 1, 1, 3, "IO19"},
    {IO_PIN, 1, 2, 4, "IO20"},		{IO_PIN, 1, 2, 5, "IO21"},
    {IO_PIN, 1, 3, 6, "IO22"},		{IO_PIN, 1, 3, 7, "IO23"},
};

int isp1016_clocks[] = {
    11, 35, 33, 0
};

int isp1016_oes[] = {
    0
};

struct ispdevinfo {
    struct isppininfo *pins;
    int		reset_pin;
    int		*clock_pins;
    int		*oe_pins;
    int		n_clockdist;
    int		n_ioclkdist;
    int		n_boxes;
    int		n_banks;
    int		boxes_per_bank;
    int		funcs_per_box;
    int		inputs_per_box;
    int		ios_per_bank;
    int		dedins_per_bank;
    int		n_input_groups;
    char	*part_ident;
} isp1016_dev = {
    isp1016_pins,
    35,
    isp1016_clocks,
    isp1016_oes,
    3,
    2,
    16,
    2,
    8,
    4,
    16,
    16,
    2,
    16,
    "ispLSI1016-80LJ"
};

struct pal isp1016 = {
    44, 32, 0, 0, 0, 0, NULL,
    isp_init, isp_pin, isp_spew, NULL, (int) &isp1016_dev
};

struct ispdecl {
    int		num;		/* pin number (-1=internal, 0=unassigned) */
    int		attr;		/* [fast] {input, output}, etc. */
    int		inindex;	/* index in input_cols */
    struct ispdecl *next;	/* next mention of this signal */
};

struct ispsig {
    struct expr	*sig;		/* signal on this pin or internal output */
    struct array *logic;	/* signal definition in array form */
    struct array *xor;		/* xor part of signal, in array form */
    int		attr;		/* [fast] {input, output}, etc. */
    int		use;
    int		flags;
    int		bank;
    int		box;
    int		input_group;
    struct ispdecl *first_decl;
    struct ctrlfn *inclock;
    struct ctrlfn *clock;
    struct ctrlfn *reset;
    struct ctrlfn *oe;
    struct ispsig *indir;
    bitmap	sigmap;
    bitmap	exclude;
    char	external_name[32];
};

/* values for flags */
#define INVERTED	1
#define PTCLOCK		2
#define PTRESET		4
#define PTOE		8
#define CLKINV		0x10
#define OEINV		0x20
#define INCLKINV	0x40
#define FIXEDBOX	0x100
#define FIXEDBANK	0x200
#define FIXEDPIN	0x400
#define EXTFEEDBACK	0x800
#define USEINVERSE	0x1000
#define IOINPUT		0x2000
#define INTFEEDBACK	0x4000

#define FAST_INPUT	(INPUT_BIT | CRIT_PATH)
#define FAST_OUTPUT	(OUTPUT_BIT | CRIT_PATH)

struct ctrlfn {
    struct expr	*exp;
    struct expr *invexp;
    struct array *logic;
    int		usebits;
    int		nterms;
    int		input_num;
    int		input_inv;
    bitmap	sigmap;
    int		bank_use;
};

struct boxinfo {
    int		n_fns;
    short	bank;
    short	box;
    int		free_fastos;
    struct ispsig *funcs[4];
    struct ctrlfn *ptoe;
    bitmap	sigmap;
    bitmap	input_groups;
};

struct bankinfo {
    int		bank;
    int		free_ios;
    struct ctrlfn *ptoe;
    struct ispsig *oeuse;
    int		oebox;
    int		n_dedins;
    struct ispsig *dedins[4];
    struct boxinfo *boxes;
};

int ext_pins[44];		/* pin number -> index in isp_decls */
struct ispdecl isp_decls[128];	/* signals mentioned in pal defn */
struct expr *input_cols[128];
struct ispsig isp_sigs[128];
struct ctrlfn ctrl_fns[32];
struct ispdevinfo *isp_dev;
struct boxinfo box_info[48];
struct bankinfo bank_info[6];
bitmap io_inputs;
bitmap fixed_io_inputs;

int n_ispdecls;
int n_inputcols;
int n_ctrlfns;
int bitmap_width;

char *signal_name();
struct ctrlfn *find_ctrlfn();
struct array *sorted_array();
struct array *isp_array();
struct array *sigmin();

static int old_errors;
static char *newline;
static int ldf_column;

#define NBITS_UNIT	32
#define NBITS_SHIFT	5
#define NBITS_MASK	(NBITS_UNIT - 1)
#define SETBIT(a, n)	((a)[(n) >> NBITS_SHIFT] |= 1 << ((n) & NBITS_MASK))
#define ISSET(a, n)	((a)[(n) >> NBITS_SHIFT] & (1 << ((n) & NBITS_MASK)))

/*
 * Initialization, called at the beginning of a pal declaration.
 */
void
isp_init()
{
    int i, j, nfastos;
    struct boxinfo *bp;
    struct bankinfo *kp;

    isp_dev = (struct ispdevinfo *) cur_pal->sp_arg;
    n_ispdecls = 0;
    n_inputcols = 0;
    n_ctrlfns = 0;
    for (i = 0; i < cur_pal->num_pins; ++i)
	ext_pins[i] = -1;
    memset(isp_decls, 0, sizeof(isp_decls));
    memset(isp_sigs, 0, sizeof(isp_sigs));
    nfastos = isp_dev->ios_per_bank / isp_dev->boxes_per_bank;
    for (i = 0, bp = box_info; i < isp_dev->n_boxes; ++i, ++bp) {
	bp->n_fns = 0;
	bp->bank = i / isp_dev->boxes_per_bank;
	bp->box = i % isp_dev->boxes_per_bank;
	bp->free_fastos = nfastos;
	bp->ptoe = NULL;
    }
    for (i = 0, kp = bank_info; i < isp_dev->n_banks; ++i, ++kp) {
	kp->bank = i;
	kp->ptoe = NULL;
	kp->free_ios = isp_dev->ios_per_bank;
	kp->n_dedins = 0;
	kp->boxes = &box_info[i * isp_dev->boxes_per_bank];
    }
    old_errors = n_errs;
}

/*
 * Process a group of pin declarations.
 */
void
isp_pin(attr, sigs, pins)
    int attr;
    struct expr *sigs, *pins;
{
    int np, i, pn, inix, rep_attr, pbit;
    struct expr *ppp, *spp;
    struct expr *sp, *pp;
    struct signal *sig;
    struct ispdecl *eip, *eiq;
    struct ispsig *isp;
    struct isppininfo *pip;
    char *signame, *type;
    struct boxinfo *bip;
    struct bankinfo *bkp;

    np = sigs->e_nel;
    ppp = NULL;
    pn = (attr & INTERNAL_SIG)? -1: 0;
    if (pins != NULL) {
	if (pins->e_nel != np) {
	    yyerror("# signals doesn't match # pins");
	} else
	    ppp = pins->e_first;
    }
    spp = sigs->e_first;
    for (; np > 0; --np, spp = spp->e_right) {
	sp = spp->e_left;
	if (ppp != NULL) {
	    pp = ppp->e_left;
	    ppp = ppp->e_right;
	    pn = pp->e_pin;
	    if (attr & INTERNAL_SIG) {
		pn /= isp_dev->funcs_per_box;
		if (pn < 0 || pn >= isp_dev->n_boxes) {
		    yyerror("internal signal location %d out of range", pn);
		    pn = -1;
		}
	    } else {
		if (pn <= 0 || pn > cur_pal->num_pins) {
		    yyerror("there is no pin %d", pn);
		    pn = 0;
		}
	    }
	}

	if (sp->e_op != SIGNAL) {
	    yyerror("not a signal");
	    continue;
	}
	sig = sp->e_signal;
	signame = signal_name(sp);
	if (attr & (INTERNAL_SIG|OUTPUT_BIT)) {
	    if ((sig->s_attr.a_bits & OUTPUT_BIT) == 0) {
		yyerror("%s is not an output", signame);
		continue;
	    }
	    if (sig->s_def[sp->e_index - sig->s_ix.i_minix] == NULL) {
		yyerror("no definition for %s", signame);
		continue;
	    }
	}

	isp = isp_sigs;
	for (inix = 0; inix < n_inputcols; ++inix, ++isp)
	    if (isp->sig->e_signal == sp->e_signal
		&& isp->sig->e_index == sp->e_index)
		break;

	eip = &isp_decls[n_ispdecls];
	eip->attr = attr;
	eip->inindex = inix;
	eip->next = NULL;

	if (inix < n_inputcols) {
	    for (eiq = isp->first_decl; eiq->next != NULL; eiq = eiq->next)
		;
	    eiq->next = eip;
	} else {
	    input_cols[inix] = sp;
	    isp->sig = sp;
	    isp->logic = NULL;
	    isp->xor = NULL;
	    isp->attr = 0;
	    isp->use = 0;
	    isp->flags = 0;
	    isp->box = -1;
	    isp->bank = -1;
	    isp->input_group = -1;
	    isp->first_decl = eip;
	    isp->inclock = NULL;
	    isp->clock = NULL;
	    isp->reset = NULL;
	    isp->oe = NULL;
	    isp->indir = NULL;
	    ++n_inputcols;
	}

	if (attr & INTERNAL_SIG) {
	    eip->num = -1;
	    if (isp->attr & INTERNAL_SIG) {
		yyerror("%s is an internal signal more than once", signame);
	    }
	    if (pn >= 0) {
		if (!add_to_box(&box_info[pn], isp, signame))
		    pn = -1;
	    }
	} else {
	    switch (attr) {
	    case CLOCKED:
		type = "clock input";
		pbit = GLOBAL_CLK;
		break;
	    case ASYNC_RESET:
		type = "global reset input";
		pbit = GLOBAL_RESET;
		break;
	    case TRISTATE:
		type = "output enable";
		pbit = GLOBAL_OE;
		break;
	    case INPUT_BIT:
		type = "input";
		pbit = INPUT_PIN;
		break;
	    case FAST_INPUT:
		type = "dedicated input";
		pbit = DED_INPUT;
		break;
	    default:
		type = "output";
		pbit = OUTPUT_PIN;
	    }
	    if (sig->s_attr.a_inclock != NULL
		&& (pbit & (GLOBAL_CLK|GLOBAL_RESET|GLOBAL_OE|DED_INPUT))) {
		yyerror("%s: %ss cannot be latched or registered",
			signame, type);
		if (attr & INPUT_BIT) {
		    attr &= ~CRIT_PATH;
		    pbit = INPUT_PIN;
		}
	    }
	    if (pn > 0) {
		pip = &isp_dev->pins[pn-1];
		if (ext_pins[pn-1] >= 0) {
		    yyerror("pin %d already used", pn);
		    pn = 0;
		} else if ((pip->bits & pbit) == 0) {
		    yyerror("pin %d is not %s %s", pn,
			    (type[0] == 'i' || type[0] == 'o'? "an": "a"),
			    type);
		    pn = 0;
		} else if (attr & INPUT_BIT) {
		    if (pip->bits & DED_INPUT) {
			bkp = &bank_info[pip->bank];
			isp->bank = pip->bank;
			bkp->dedins[bkp->n_dedins++] = isp;
			isp->attr |= CRIT_PATH;
		    }
		} else if (attr & OUTPUT_BIT) {
		    if (attr & CRIT_PATH) {
			bkp = &bank_info[pip->bank];
			bip = &bkp->boxes[pip->box];
			if (!add_to_box(bip, isp, signame))
			    pn = 0;
		    } else if ((isp->flags & FIXEDBANK)
			       && pip->bank != isp->bank) {
			yyerror("%s is already fixed in GLB bank %c",
				signame, isp->bank + 'A');
			pn = 0;
		    } else {
			isp->flags |= FIXEDBANK;
			isp->bank = pip->bank;
		    }
		}
	    }
	    if (pn != 0) {
		ext_pins[pn-1] = n_ispdecls;
		isp->flags |= FIXEDPIN;
	    }
	    eip->num = pn;
	    rep_attr = attr & isp->attr;
	    if (rep_attr & OUTPUT_BIT) {
		yyerror("%s is declared as an output more than once", signame);
	    }
	    if (rep_attr & INPUT_BIT) {
		yyerror("%s is declared as an input more than once", signame);
	    }
	}
	isp->attr |= attr;
	++n_ispdecls;
	if (isp->attr & (OUTPUT_BIT|INTERNAL_SIG) && isp->attr & INPUT_BIT) {
	    yyerror("%s is declared as both an input and an output", signame);
	}
    }
}

/*
 * Process the pal and output LDF for it.
 */
void
isp_spew()
{
    if (n_errs > old_errors)
	return;
    make_external_names();
    conv_inputs(input_cols, n_inputcols);
    isp_logic();
    allocate_ctrlfns();
    add_internal();
    decide_feedback();
    make_bitmaps();
    if (n_errs > old_errors)
	return;
    allocate_boxes();
    allocate_io_pins();
/*    if (pdebug)
	isp_dump(); */
    if (n_errs > old_errors)
	return;
    dump_ldf();
}

/*
 * Assign a signal to a particular box, if possible.
 */
int
add_to_box(bip, isp, signame)
    struct boxinfo *bip;
    struct ispsig *isp;
    char *signame;
{
    if (isp->flags & FIXEDBOX) {
	if (bip->bank != isp->bank || bip->box != isp->box) {
	    yyerror("%s is already fixed to GLB %c%d",
		    signame, isp->bank + 'A', isp->box);
	    return 0;
	}
	return 1;
    }
    if (isp->flags & FIXEDBANK && bip->bank != isp->bank) {
	yyerror("%s is already fixed to GLB bank %c",
		signame, isp->bank + 'A');
	return 0;
    }
    if (bip->n_fns >= isp_dev->funcs_per_box) {
	yyerror("%s: GLB %c%d is already full", signame,
		bip->bank + 'A', bip->box);
	return 0;
    }
    isp->flags |= FIXEDBOX | FIXEDBANK;
    isp->bank = bip->bank;
    isp->box = bip->box;
    bip->funcs[bip->n_fns] = isp;
    ++bip->n_fns;
    return 1;
}

/*
 * Generate an appropriate name to go in the LDF file for
 * each signal.
 */
make_external_names()
{
    struct ispsig *isp;
    struct signal *sig;
    int xxx;

    xxx = 0;
    for (isp = isp_sigs; isp < &isp_sigs[n_inputcols]; ++isp) {
	if (isp->sig->e_op == SIGNAL) {
	    sig = isp->sig->e_signal;
	    if (sig->s_ix.i_nel == 1 && sig->s_ix.i_minix == 0) {
		sprintf(isp->external_name, "%s%.28s",
			(sig->s_attr.a_bits & LOW_TRUE)? "_": "",
			sig->s_ident->i_name);
	    } else {
		sprintf(isp->external_name, "%s%.24s_%d",
			(sig->s_attr.a_bits & LOW_TRUE)? "_": "",
			sig->s_ident->i_name, isp->sig->e_index);
	    }
	} else {
	    sprintf(isp->external_name, "__XXX_%d", xxx++);
	}
    }
}

/*
 * Make sure there is an internal output corresponding to each
 * output pin and perform various other checks.
 */
add_internal()
{
    struct ispsig *isp, *isi;
    struct ispdecl *eip, *lip;
    int i, inn, inv;
    struct boxinfo *bip;

    for (isp = isp_sigs; isp < &isp_sigs[n_inputcols]; ++isp) {

	switch (isp->attr & (OUTPUT_BIT|INTERNAL_SIG)) {
	case INTERNAL_SIG:
	    if (isp->sig->e_signal->s_attr.a_bits & TRISTATE)
		yyerror("warning: internal signal %s cannot be tristate",
			signal_name(isp->sig));
	    break;

	case OUTPUT_BIT:
	    /*
	     * For an output signal, check if it is just a copy
	     * of an existing internal signal.  If so, we don't
	     * need an internal signal to generate it.  But the
	     * output does constrain which bank (or box) the
	     * internal signal can go in.
	     */
	    isp->flags |= EXTFEEDBACK;
	    if (isp->xor == NULL && isp->clock == NULL && isp->reset == NULL
		&& single_input(isp->logic, &inn, &inv)) {
		isi = &isp_sigs[inn];
		if ((isi->attr & (INTERNAL_SIG | OUTPUT_BIT)) == INTERNAL_SIG
		    && !(isp->flags & FIXEDBANK && isi->flags & FIXEDBANK
			  && isp->bank != isi->bank
			 || isp->flags & FIXEDBOX && isi->flags & FIXEDBOX)) {
		    isp->indir = isi;
		    isi->attr |= OUTPUT_BIT;
		    isi->flags |= isp->flags & (FIXEDBANK | FIXEDBOX | PTOE)
			| INTFEEDBACK;
		    isi->oe = isp->oe;
		    if (isp->flags & FIXEDBANK)
			isi->bank = isp->bank;
		    if (isp->flags & FIXEDBOX) {
			isi->box = isp->box;
			/* replace isp with isi in the box function list. */
			bip = &bank_info[isp->bank].boxes[isp->box];
			for (i = 0; i < bip->n_fns; ++i)
			    if (bip->funcs[i] == isp)
				bip->funcs[i] = isi;
		    }
		    break;
		}
	    }

	    for (lip = isp->first_decl; lip->next != NULL; lip = lip->next)
		;
	    eip = &isp_decls[n_ispdecls++];
	    lip->next = eip;
	    eip->num = -1;
	    eip->attr = INTERNAL_SIG;
	    eip->inindex = lip->inindex;
	    eip->next = NULL;
	    isp->attr |= INTERNAL_SIG;
	    break;

	case INTERNAL_SIG|OUTPUT_BIT:
	    if ((isp->flags & INTFEEDBACK) == 0)
		isp->flags |= EXTFEEDBACK;
	    break;
	}
    }
}

/*
 * Convert the definitions for all internal and control signals
 * to logic functions.
 */
isp_logic()
{
    struct ispsig *isp;
    struct signal *sig;
    struct expr *def;
    int pol, n;

    if (pdebug > 1) {
	printf("Input columns:\n");
	for (n = 0; n < n_inputcols; ++n)
	    printf("%d = %s\n", n, signal_name(isp_sigs[n].sig));
    }

    for (isp = isp_sigs; isp < &isp_sigs[n_inputcols]; ++isp) {
	sig = isp->sig->e_signal;

	if (isp->attr & OUTPUT_BIT && sig->s_attr.a_enable != NULL)
	    isp->oe = find_ctrlfn(sig->s_attr.a_enable, TRISTATE,
				  &isp->flags);

	if (isp->attr & (INTERNAL_SIG | OUTPUT_BIT)) {
	    if (sig->s_attr.a_clock != NULL)
		isp->clock = find_ctrlfn(sig->s_attr.a_clock, CLOCKED,
					 &isp->flags);
	    if (sig->s_attr.a_reset != NULL)
		isp->reset = find_ctrlfn(sig->s_attr.a_reset, ASYNC_RESET,
					 &isp->flags);

	    def = sig->s_def[isp->sig->e_index - sig->s_ix.i_minix];
	    if (def != NULL) {
		if (pdebug > 1) {
		    printf("%s = ", signal_name(isp->sig));
		    pexpr(def);
		    printf("\n");
		}
		isp->logic = isp_array(def, &sig->s_attr, &isp->xor, &pol);
		if (pdebug > 1) {
		    printf("array form%s\n", pol==Z? " (inverted)": "");
		    prary(isp->logic, stdout);
		    if (isp->xor != NULL) {
			printf("XOR\n");
			prary(isp->xor, NULL);
		    }
		}
		note_uses(isp->logic, INPUT_BIT);
		if (isp->xor != NULL)
		    note_uses(isp->xor, INPUT_BIT);
		if (pol == Z)
		    isp->flags |= INVERTED;
	    } else
		isp->logic = isp->xor = NULL;
	}
    }

    /*
     * Now check for I/O pins with input clock signals
     * that are actually used as inputs.
     */
    do {
	n = 0;
	for (isp = isp_sigs; isp < &isp_sigs[n_inputcols]; ++isp) {
	    sig = isp->sig->e_signal;
	    if (isp->attr & (INPUT_BIT|OUTPUT_BIT) && isp->use != 0
		&& sig->s_attr.a_inclock != NULL && isp->inclock == NULL) {
		isp->inclock = find_ctrlfn(sig->s_attr.a_inclock, INPUT_CLOCK,
					   &isp->flags);
		++n;
	    }
	}
    } while (n > 0);
}

/*
 * Allocate clock and reset inputs.
 * Work out whether to use product-term clock/reset functions
 * or the distribution networks.
 */
allocate_ctrlfns()
{
    struct ctrlfn *cfp;
    struct ispsig *isp;
    int bits;

    /*
     * Allocate the reset input.
     */
    alloc_ctrl_pins(&isp_dev->reset_pin, 1, ASYNC_RESET, "reset");

    /*
     * Work out which functions use the PT reset.
     */
    for (isp = isp_sigs; isp < &isp_sigs[n_inputcols]; ++isp) {
	if (isp->reset != NULL) {
	    cfp = isp->reset;
	    if (cfp->input_num < 0 || cfp->input_inv
		|| (isp_sigs[cfp->input_num].attr & ASYNC_RESET) == 0)
		isp->flags |= PTRESET;
	}
    }

    /*
     * Allocate clock inputs.
     */
    bits = CLOCKED | CRIT_PATH;
    do {
	alloc_ctrl_pins(isp_dev->clock_pins, 0, bits, "clock");
	bits ^= CRIT_PATH;
    } while ((bits & CRIT_PATH) == 0);

    /*
     * Work out which functions use a PT clock.
     * XXX we should allow for internal signals generated in the clock
     * GLB to be used via the clock distribution network.
     */
    for (isp = isp_sigs; isp < &isp_sigs[n_inputcols]; ++isp) {
	if (isp->clock != NULL) {
	    cfp = isp->clock;
	    if (cfp->input_num < 0
		|| (isp_sigs[cfp->input_num].attr & CLOCKED) == 0)
		isp->flags |= PTCLOCK;
	}
    }

    /*
     * Allocate OE controls.
     */
    alloc_ctrl_pins(isp_dev->oe_pins, 0, GLOBAL_OE, "output enable");

    /*
     * Work out which functions use a PT output enable.
     */
    for (isp = isp_sigs; isp < &isp_sigs[n_inputcols]; ++isp) {
	if (isp->oe != NULL) {
	    cfp = isp->oe;
	    if (cfp->input_num < 0
		|| (isp_sigs[cfp->input_num].attr & TRISTATE) == 0)
		isp->flags |= PTOE;
	}
    }
}

/*
 * Allocate pins from `list' for unassigned pins with
 * attributes matching `attr'.
 */
alloc_ctrl_pins(list, nlist, attr, what)
    int *list, nlist, attr;
    char *what;
{
    struct ispdecl *ipd;
    int i, n;

    for (ipd = isp_decls; ipd < &isp_decls[n_ispdecls]; ++ipd) {
	if (ipd->num == 0 && (ipd->attr & attr) == attr) {
	    for (i = 0; (nlist == 0 || i < nlist) && (n = list[i]) > 0; ++i) {
		if (ext_pins[n-1] < 0) {
		    assign_pin(ipd, n);
		    break;
		}
	    }
	    if (ipd->num == 0) {
		yyerror("no available %s pin for %s", what,
			signal_name(isp_sigs[ipd->inindex].sig));
	    }
	}
    }
}

/*
 * Set up bitmaps indicating which signals are used.
 */
make_bitmaps()
{
    struct ctrlfn *cfp;
    struct ispsig *isp, *isi;
    struct boxinfo *bip;
    int i, ii;
    struct ispdecl *ipd;
    bitmap xmap, igroups;

    bitmap_width = (n_inputcols + NBITS_UNIT) >> NBITS_SHIFT;
    for (cfp = ctrl_fns; cfp < &ctrl_fns[n_ctrlfns]; ++cfp) {
	memset(cfp->sigmap, 0, sizeof(bitmap));
	make_bitmap(cfp->sigmap, cfp->logic);
    }
    memset(io_inputs, 0, sizeof(bitmap));
    memset(fixed_io_inputs, 0, sizeof(bitmap));
    for (isp = isp_sigs, ii = 0; ii < n_inputcols; ++ii, ++isp) {
	memset(isp->sigmap, 0, sizeof(bitmap));
	memset(isp->exclude, 0, sizeof(bitmap));
	if (isp->attr & INTERNAL_SIG) {
	    make_bitmap(isp->sigmap, isp->logic);
	    make_bitmap(isp->sigmap, isp->xor);
	    if (isp->flags & PTCLOCK)
		add_bits(isp->sigmap, isp->clock->sigmap);
	    if (isp->flags & PTRESET)
		add_bits(isp->sigmap, isp->reset->sigmap);
	}
	if ((isp->flags & EXTFEEDBACK)
	    || (isp->attr & (INPUT_BIT|CRIT_PATH)) == INPUT_BIT) {
	    isp->flags |= IOINPUT;
	    SETBIT(io_inputs, ii);
	    for (ipd = isp->first_decl; ipd != NULL; ipd = ipd->next) {
		if ((ipd->attr & (INPUT_BIT|OUTPUT_BIT)) && ipd->num > 0) {
		    isp->input_group = isp_dev->pins[ipd->num - 1].group;
		    SETBIT(fixed_io_inputs, ii);
		    break;
		}
	    }
	}
    }
    for (isp = isp_sigs; isp < &isp_sigs[n_inputcols]; ++isp) {
	if ((isp->attr & INTERNAL_SIG) == 0)
	    continue;
	if (!intersect_maps(xmap, isp->sigmap, io_inputs))
	    continue;
	memset(igroups, 0, sizeof(bitmap));
	for (ii = -1; (ii = next_bit(xmap, ii)) >= 0; ) {
	    isi = &isp_sigs[ii];
	    add_bits(isi->exclude, xmap);
	    if (isi->input_group >= 0) {
		if (ISSET(igroups, isi->input_group)) {
		    yyerror("%s uses two inputs from the same group",
			    signal_name(isp->sig));
		    yyerror("(condition detected for input %s)",
			    signal_name(isi->sig));
		} else {
		    SETBIT(igroups, isi->input_group);
		}
	    }
	}
    }
    for (bip = box_info; bip < &box_info[isp_dev->n_boxes]; ++bip) {
	memset(bip->sigmap, 0, sizeof(bitmap));
	memset(bip->input_groups, 0, sizeof(bitmap));
	for (i = 0; i < bip->n_fns; ++i) {
	    isp = bip->funcs[i];
	    if (input_conflict(bip, isp->sigmap, xmap))
		yyerror("input conflict putting %s in GLB %c%d",
			signal_name(isp->sig), bip->bank + 'A', bip->box);
	    add_bits(bip->sigmap, isp->sigmap);
	    add_bits(bip->input_groups, xmap);
	}
    }
}

/*
 * Test whether the new inputs for signal isp would conflict with the
 * inputs already in use by box bip.
 */
int
input_conflict(bip, inputs, newgroups)
    struct boxinfo *bip;
    unsigned long *inputs, *newgroups;
{
    bitmap xmap;
    struct ispsig *isi;
    int ii, ret;

    ret = 0;
    memset(newgroups, 0, sizeof(bitmap));
    if (intersect_maps(xmap, inputs, fixed_io_inputs)
	&& remove_bits(xmap, bip->sigmap)) {
	for (ii = -1; (ii = next_bit(xmap, ii)) >= 0; ) {
	    isi = &isp_sigs[ii];
	    if (ISSET(bip->input_groups, isi->input_group))
		ret = 1;
	    else
		SETBIT(newgroups, isi->input_group);
	}
    }
    return ret;
}

/*
 * Add bits for the inputs used in `ap' to the bitmap `bm'.
 */
make_bitmap(bm, ap)
    unsigned long *bm;
    struct array *ap;
{
    struct cube *cp;
    int i;

    if (ap == NULL)
	return;
    for (cp = ap->a_first; cp != NULL; cp = cp->c_next) {
	for (i = 0; i < n_inputcols; ++i) {
	    if (cp->c_logic[i] == Z || cp->c_logic[i] == O)
		SETBIT(bm, i);
	}
    }
}

/*
 * Allocate functions to boxes.
 */
allocate_boxes()
{
    int i, pass, bank, bit, do_empties;
    struct ispsig *isp, *isi;
    struct boxinfo *bp, *lastbox, *bestbox;
    struct ctrlfn *cfp, *oef;
    struct bankinfo *kp;
    bitmap ded_inputs;
    bitmap sig_dedin_map;
    bitmap xmap;
    struct ispsig *new_dedins[4];
    int num_dedins, bank_max_dedins;

    /*
     * Make a list of the dedicated inputs and a list of the
     * I/O pins used as inputs or feedback.
     */
    memset(ded_inputs, 0, sizeof(bitmap));
    isi = isp_sigs;
    for (i = 0; i < n_inputcols; ++i, ++isi) {
	if ((isi->attr & (INPUT_BIT|CRIT_PATH)) == FAST_INPUT)
	    SETBIT(ded_inputs, i);
    }

    /*
     * Update the number of free I/O pins on each bank
     * by looking at the signals which are fixed to some bank.
     */
    for (isp = isp_sigs; isp < &isp_sigs[n_inputcols]; ++isp) {
	if (isp->bank >= 0
	    && ((isp->attr & (INTERNAL_SIG|OUTPUT_BIT))
		== (INTERNAL_SIG|OUTPUT_BIT)
		|| (isp->attr & (INPUT_BIT|CRIT_PATH)) == INPUT_BIT)) {
	    kp = &bank_info[isp->bank];
	    if (kp->free_ios > 0)
		--kp->free_ios;
	    else {
		yyerror("%s: too many I/Os for bank %c",
			signal_name(isp->sig), isp->bank + 'A');
	    }
	    if ((isp->attr & (OUTPUT_BIT|CRIT_PATH)) == FAST_OUTPUT
		&& isp->box >= 0) {
		bp = &kp->boxes[isp->box];
		if (--bp->free_fastos < 0)
		    yyerror("%s: too many fast outputs for GLB %c%d",
			    signal_name(isp->sig), bp->bank + 'A', bp->box);
	    }
	}
    }

    /*
     * Look at OE requirements of outputs that are already fixed
     * to a specific bank.
     */
    for (cfp = ctrl_fns; cfp < &ctrl_fns[n_ctrlfns]; ++cfp)
	cfp->bank_use = 0;
    for (isp = isp_sigs; isp < &isp_sigs[n_inputcols]; ++isp) {
	if ((isp->attr & OUTPUT_BIT) == 0 || isp->oe == NULL
	    || (isp->flags & (FIXEDBANK|PTOE)) != (FIXEDBANK|PTOE)
	    || (isp->oe->bank_use & (1 << isp->bank)) != 0)
	    continue;
	kp = &bank_info[isp->bank];
	cfp = isp->oe;
	if (kp->ptoe == NULL) {
	    kp->ptoe = cfp;
	    kp->oeuse = isp;
	    cfp->bank_use |= 1 << isp->bank;
	} else {
	    yyerror("%s: too many output enable functions for GLB bank %c",
		    signal_name(isp->sig), isp->bank + 'A');
	    yyerror("(bank has OE function for signal %s)",
		    signal_name(kp->oeuse->sig));
	}
    }

    /*
     * Allocate each unallocated internal signal to the first box
     * that it can go in and that it fits in.  If a signal has an
     * output enable and is not fixed to a specific bank, we first
     * try to put it in a bank which already has that OE function.
     * We do the signals which are fixed to some bank first, then
     * signals which use a dedicated input, then the rest.
     */
    lastbox = &box_info[isp_dev->n_boxes];
    for (pass = 2; pass >= 0; --pass) {
	if (pass == 0) {
	    /*
	     * Try to allocate the OE functions to boxes before
	     * the last pass.
	     */
	    for (kp = bank_info; kp < &bank_info[isp_dev->n_banks]; ++kp) {
		kp->oebox = -1;
		if (kp->ptoe == NULL)
		    continue;
		bp = kp->boxes;
		for (i = 0; i < isp_dev->boxes_per_bank; ++i, ++bp) {
		    if (count_sum_bits(kp->ptoe->sigmap, bp->sigmap, ded_inputs
				       ) < isp_dev->inputs_per_box
			&& !input_conflict(bp, kp->ptoe->sigmap, xmap)) {
			kp->oebox = i;
			bp->ptoe = kp->ptoe;
			add_bits(bp->sigmap, kp->ptoe->sigmap);
			add_bits(bp->input_groups, xmap);
			break;
		    }
		}
	    }
	}
	for (isp = isp_sigs; isp < &isp_sigs[n_inputcols]; ++isp) {
	    if ((isp->attr & INTERNAL_SIG) == 0 || isp->box >= 0)
		continue;
	    bank = isp->bank;
	    num_dedins = 0;
	    bank_max_dedins = isp_dev->dedins_per_bank;
	    if (intersect_maps(sig_dedin_map, isp->sigmap, ded_inputs)) {
		bit = -1;
		while ((bit = next_bit(sig_dedin_map, bit)) >= 0) {
		    isi = &isp_sigs[bit];
		    if (isi->bank >= 0) {
			if (bank >= 0) {
			    if (bank != isi->bank && pass == 0)
				yyerror("%s: dedicated input bank conflict",
					signal_name(isp->sig));
			} else
			    bank = isi->bank;
		    } else {
			if (--bank_max_dedins >= 0)
			    new_dedins[num_dedins++] = isi;
		    }
		}
		if (bank >= 0 && bank_info[bank].n_dedins > bank_max_dedins
		    || bank_max_dedins < 0) {
		    yyerror("%s: too many dedicated inputs\n",
			    signal_name(isp->sig));
		    bank_max_dedins = 0;
		}
	    }
	    if (bank < 0 && (pass == 2 || pass == 1 && num_dedins == 0))
		continue;
	    oef = ((isp->attr & OUTPUT_BIT) && (isp->flags & PTOE))?
		isp->oe: NULL;
	    bestbox = NULL;
	    for (do_empties = 0; do_empties <= 1; ++do_empties) {
		for (bp = box_info; bp < lastbox; ++bp) {
		    if (do_empties != (bp->n_fns == 0))
			continue;
		    if ((bank < 0 || bp->bank == bank) && can_assign(isp, bp)
			&& count_sum_bits(isp->sigmap, bp->sigmap, ded_inputs
					  ) < isp_dev->inputs_per_box
			&& bank_info[bp->bank].n_dedins <= bank_max_dedins
			&& !input_conflict(bp, isp->sigmap, xmap)) {
			if (oef == NULL
			    || (oef->bank_use & (1 << bp->bank)) != 0) {
			    bestbox = bp;
			    break;
			}
			if (bestbox == NULL)
			    bestbox = bp;
		    }
		}
		if (bestbox != NULL)
		    break;
	    }
	    if (bestbox != NULL) {
		isp->bank = bestbox->bank;
		isp->box = bestbox->box;
		bestbox->funcs[bestbox->n_fns++] = isp;
		input_conflict(bestbox, isp->sigmap, xmap);
		add_bits(bestbox->input_groups, xmap);
		add_bits(bestbox->sigmap, isp->sigmap);
		if ((isp->attr & (OUTPUT_BIT|CRIT_PATH)) == FAST_OUTPUT)
		    --bestbox->free_fastos;
		kp = &bank_info[bestbox->bank];
		if (oef != NULL && kp->ptoe == NULL) {
		    kp->ptoe = oef;
		    kp->oeuse = isp;
		    oef->bank_use |= 1 << bestbox->bank;
		}
		if ((isp->attr & OUTPUT_BIT) && (isp->flags & FIXEDBANK) == 0)
		    --kp->free_ios;
		for (i = 0; i < num_dedins; ++i) {
		    isi = new_dedins[i];
		    kp->dedins[kp->n_dedins++] = isi;
		    isi->bank = kp->bank;
		    if (pdebug)
			printf("%s allocated to bank %c\n",
			       signal_name(isi->sig), kp->bank + 'A');
		}
		if (pdebug)
		    printf("%s allocated to GLB %c%d\n",
			   signal_name(isp->sig), isp->bank + 'A', isp->box);
	    } else if (pass == 0) {
		yyerror("couldn't allocate %s to a GLB",
			signal_name(isp->sig));
	    }
	}
    }

    /*
     * For each bank, find a GLB for its OE function to go in,
     * if it doesn't have one already.
     */
    for (kp = bank_info; kp < &bank_info[isp_dev->n_banks]; ++kp) {
	if (kp->ptoe == NULL || kp->oebox >= 0)
	    continue;
	bp = kp->boxes;
	for (i = 0; i < isp_dev->boxes_per_bank; ++i, ++bp) {
	    if (count_sum_bits(kp->ptoe->sigmap, bp->sigmap, ded_inputs
			       ) < isp_dev->inputs_per_box
		&& !input_conflict(bp, kp->ptoe->sigmap, xmap)) {
		kp->oebox = i;
		bp->ptoe = kp->ptoe;
		add_bits(bp->sigmap, kp->ptoe->sigmap);
		add_bits(bp->input_groups, xmap);
		break;
	    }
	}
	if (kp->oebox < 0)
	    yyerror("couldn't find GLB for OE function in bank %c",
		    kp->bank + 'A');
    }

}

/*
 * Assign pin numbers to any remaining unassigned I/O pins.
 */
allocate_io_pins()
{
    int i;
    struct ispdecl *ipd;
    struct ispsig *isp;
    struct isppininfo *pip;
    struct bankinfo *kp;

    for (ipd = isp_decls; ipd < &isp_decls[n_ispdecls]; ++ipd) {
	if (ipd->num != 0)
	    continue;
	if (ipd->attr & OUTPUT_BIT) {
	    isp = &isp_sigs[ipd->inindex];
	    if (isp->bank >= 0) {
		pip = isp_dev->pins;
		for (i = 1; i <= cur_pal->num_pins; ++i, ++pip) {
		    if (ext_pins[i-1] < 0 && (pip->bits & OUTPUT_PIN)
			&& pip->bank == isp->bank) {
			assign_pin(ipd, i);
			break;
		    }
		}
		if (ipd->num == 0) {
		    yyerror("no available I/O pin for %s in bank %c",
			    signal_name(isp->sig), isp->bank + 'A');
		}
	    }
	} else if (ipd->attr & INPUT_BIT) {
	    isp = &isp_sigs[ipd->inindex];
	    pip = isp_dev->pins;
	    for (i = 1; i <= cur_pal->num_pins; ++i, ++pip) {
		if (ext_pins[i-1] >= 0 || (pip->bits & INPUT_PIN) == 0
		    || (isp->bank >= 0 && isp->bank != pip->bank)
		    || (pip->bits & IO_PIN)
		       != (isp->attr & CRIT_PATH? INPUT_PIN: IO_PIN))
		    continue;
		if (isp->bank < 0) {
		    kp = &bank_info[pip->bank];
		    if (ipd->attr & CRIT_PATH) {
			if (kp->n_dedins < isp_dev->dedins_per_bank)
			    kp->dedins[kp->n_dedins++] = isp;
			else
			    continue;
		    } else {
			if (kp->free_ios > 0)
			    --kp->free_ios;
			else
			    continue;
		    }
		}
		assign_pin(ipd, i);
		break;
	    }
	    if (ipd->num == 0) {
		yyerror("no available input pin for %s\n",
			signal_name(isp->sig));
	    }
	}
    }
}

/*
 * Assign a pin number to a declaration.
 */
assign_pin(ipd, n)
    struct ispdecl *ipd;
    int n;
{
    ext_pins[n-1] = ipd - isp_decls;
    ipd->num = n;
    if (pdebug)
	printf("%s assigned to pin %d\n",
	       signal_name(isp_sigs[ipd->inindex].sig), n);
}

/*
 * Decide on feedback paths and signal polarities.
 */
decide_feedback()
{
    struct ispsig *isp;

    for (isp = isp_sigs; isp < &isp_sigs[n_inputcols]; ++isp) {
	if (isp->attr & (INTERNAL_SIG | OUTPUT_BIT)) {
	    if ((isp->use & INPUT_BIT) == 0 || isp->oe == NULL)
		isp->flags &= ~EXTFEEDBACK;
	    if (isp->flags & EXTFEEDBACK) {
		if (isp->sig->e_signal->s_attr.a_bits & LOW_TRUE)
		    isp->flags |= USEINVERSE;
		isp->indir = NULL;
	    } else {
		if (isp->flags & INVERTED)
		    isp->flags |= USEINVERSE;
	    }
	} else {
	    if (isp->sig->e_signal->s_attr.a_bits & LOW_TRUE)
		isp->flags |= USEINVERSE;
	}
    }
}

/*
 * Output an LDF file for the device.
 */
dump_ldf()
{
    int i, n, p;
    struct ispdecl *ipd;
    struct boxinfo *bp;
    time_t t;
    char buf[64];

    strcat(pal_fname, ".ldf");
    if ((pal_out = fopen(pal_fname, "w")) == NULL) {
	yyerror("can't create %s", pal_fname);
	return;
    }

    newline = jedec_cr? "\r\n": "\n";
    time(&t);
    strftime(buf, sizeof(buf), "%A %e %b %Y %X", localtime(&t));
    fprintf(pal_out, "// %s generated on %s by logic%s",
	    pal_fname, buf, newline);
    fprintf(pal_out, "%sLDF 1 DESIGNLDF;%sDESIGN %s;%s%s", newline,
	    newline, pal_ident, newline, newline);

    fprintf(pal_out, "PART %s;%s%s", isp_dev->part_ident, newline, newline);

    for (i = 0; (n = isp_dev->clock_pins[i]) > 0; ++i)
	if (n == isp_dev->reset_pin) {
	    p = ext_pins[n-1];
	    fprintf(pal_out, "OPTION %s_As_Reset O%s;%s",
		    isp_dev->pins[n-1].ioc_name,
		    (p < 0 || (isp_decls[p].attr & CLOCKED) == 0)? "N": "FF",
		    newline);
	    break;
	}

    fprintf(pal_out, "DECLARE%sEND;%s%s", newline, newline, newline);

    for (bp = box_info; bp < &box_info[isp_dev->n_boxes]; ++bp) {
	if (bp->n_fns > 0 || bp->ptoe != NULL)
	    isp_declare_glb(bp);
    }

    for (ipd = isp_decls; ipd < &isp_decls[n_ispdecls]; ++ipd) {
	if (ipd->num > 0 && (ipd->attr & ASYNC_RESET) == 0)
	    isp_declare_pin(ipd);
    }

    fprintf(pal_out, "END;%s", newline);
    fclose(pal_out);
}

/*
 * Output the declaration for an I/O pin to the LDF file.
 */
isp_declare_pin(ipd)
    struct ispdecl *ipd;
{
    struct isppininfo *pip;
    struct ispsig *isp;
    char *type, *suffix;
    int d;

    pip = &isp_dev->pins[ipd->num - 1];
    isp = &isp_sigs[ipd->inindex];
    fprintf(pal_out, "SYM IOC %s 1;%s", pip->ioc_name, newline);
    if (ipd->attr & CLOCKED)
	type = "CLK";
    else if (pip->bits & OUTPUT_PIN)
	type = "IO";
    else
	type = "I";
    if (ipd->attr & CLOCKED)
	suffix = "_CLK";
    else if (ipd->attr & ASYNC_RESET)
	suffix = "_RE";
    else if (ipd->attr & TRISTATE)
	suffix = "_OE";
    else
	suffix = "";
    fprintf(pal_out, "XPIN %s X%s%s", type, isp->external_name, suffix);
    if (isp->flags & FIXEDPIN)
	fprintf(pal_out, " LOCK %d", ipd->num);
    if ((ipd->attr & (OUTPUT_BIT|CRIT_PATH)) == FAST_OUTPUT)
	fprintf(pal_out, " CRITICAL");
    fprintf(pal_out, ";%s", newline);
    if ((pip->bits & OUTPUT_PIN) == 0 || (ipd->attr & INPUT_BIT)) {
	if (isp->inclock != NULL) {
	    fprintf(pal_out, "I%c%c1(%s, X%s, ",
		    isp->sig->e_signal->s_attr.a_bits & INPUT_LATCH? 'L': 'D',
		    invert_ctrlfn(isp, INPUT_CLOCK)? '2': '1',
		    isp->external_name, isp->external_name);
	    ldf_ctrlfn(isp->inclock, CLOCKED);
	} else {
	    fprintf(pal_out, "IB11(%s%s, X%s%s", isp->external_name, suffix,
		    isp->external_name, suffix);
	}
    } else {
	d = '1';
	if ((isp->sig->e_signal->s_attr.a_bits & LOW_TRUE)
	    != (isp->flags & INVERTED? LOW_TRUE: 0))
	    ++d;
	if (invert_ctrlfn(isp, TRISTATE))
	    d += 2;
	if (invert_ctrlfn(isp, INPUT_CLOCK))
	    d += 4;
	if ((isp->flags & EXTFEEDBACK) == 0) {
	    fprintf(pal_out, "O%c%c1(X%s, ", (isp->oe != NULL? 'T': 'B'),
		    d, isp->external_name);
	    if (isp->attr & INTERNAL_SIG)
		fprintf(pal_out, "%s", isp->external_name);
	    else
		isp_plogic(isp->logic, NULL);
	} else {
	    if (isp->inclock == NULL)
		suffix = "";
	    else if (isp->sig->e_signal->s_attr.a_bits & INPUT_LATCH)
		suffix = "IL";
	    else
		suffix = "ID";
	    fprintf(pal_out, "BI%s%c1(%s, X%s, ", suffix, d,
		    isp->external_name, isp->external_name);
	    if (isp->attr & INTERNAL_SIG)
		fprintf(pal_out, "%s_O", isp->external_name);
	    else
		isp_plogic(isp->logic, NULL);
	    if (isp->inclock != NULL) {
		fprintf(pal_out, ", ");
		ldf_ctrlfn(isp->inclock, CLOCKED);
	    }
	}
	if (isp->oe != NULL) {
	    if (isp->flags & PTOE)
		fprintf(pal_out, ", __OE_%c", isp->bank + 'A');
	    else {
		fprintf(pal_out, ", ");
		ldf_ctrlfn(isp->oe, TRISTATE);
	    }
	}
    }
    fprintf(pal_out, ");%sEND;%s%s", newline, newline, newline);
}

/*
 * Output the declaration for a GLB to the LDF file.
 */
isp_declare_glb(bp)
    struct boxinfo *bp;
{
    int i, doneclk, start_col;
    struct ispsig *isp;
    char *suffix;

    if (verbose) {
	struct expr *sigs[4], *ep, *eq;
	struct signal *sig;
	int i, j, n;

	printf("GLB %c%d", bp->bank + 'A', bp->box);
	for (i = 0; i < bp->n_fns; ++i)
	    sigs[i] = bp->funcs[i]->sig;
	suffix = ":";
	for (i = 0; i < bp->n_fns; ++i) {
	    if ((ep = sigs[i]) == NULL)
		continue;
	    sig = ep->e_signal;
	    printf("%s %s", suffix, sig->s_ident->i_name);
	    if (!(sig->s_ix.i_nel == 1 && sig->s_ix.i_minix == 0)) {
		printf("[%d", ep->e_index);
		n = 0;
		for (j = i+1; j < bp->n_fns; ++j) {
		    if ((eq = sigs[j]) != NULL && eq->e_signal == sig) {
			if (n >= 0 && eq->e_index == ep->e_index + n + 1)
			    ++n;
			else if (n <= 0 && eq->e_index == ep->e_index + n - 1)
			    --n;
			else {
			    if (n != 0) {
				printf(":%d", ep->e_index + n);
				n = 0;
			    }
			    printf(",%d", eq->e_index);
			    ep = eq;
			}
			sigs[j] = NULL;
		    }
		}
		if (n != 0)
		    printf(":%d", ep->e_index + n);
		printf("]");
	    }
	    suffix = ",";
	}
	if (bp->ptoe != NULL)
	    printf("%s bank OE", suffix);
	for (i = 0; i < bp->n_fns; ++i) {
	    int npt = 0;
	    struct cube *cp;
	    isp = bp->funcs[i];
	    for (cp = isp->logic->a_first; cp != NULL; cp = cp->c_next)
		if (cp->c_logic[n_inputcols] == O)
		    ++npt;
	    printf("%s%d", (i==0? " (": ","), npt);
	}
	printf(" PT)\n");
    }
    fprintf(pal_out, "SYM GLB %c%d 1;%s", bp->bank + 'A', bp->box, newline);
    for (i = 0; i < bp->n_fns; ++i) {
	isp = bp->funcs[i];
	suffix = (isp->flags & EXTFEEDBACK)? "_O": "";
	fprintf(pal_out, "SIGTYPE %s%s ", isp->external_name, suffix);
	if (isp->sig->e_signal->s_attr.a_bits & CRIT_PATH)
	    fprintf(pal_out, "CRITICAL ");
	fprintf(pal_out, "%s ", (isp->clock != NULL)? "REG": "ASYNC");
	fprintf(pal_out, "OUT NOMIN;%s", newline);
    }
    if (bp->ptoe != NULL) {
	fprintf(pal_out, "SIGTYPE __OE_%c OE;%s", bp->bank + 'A', newline);
    }
    fprintf(pal_out, "EQUATIONS%s", newline);
    doneclk = 0;
    for (i = 0; i < bp->n_fns; ++i) {
	isp = bp->funcs[i];
	suffix = (isp->flags & EXTFEEDBACK)? "_O": "";
	fprintf(pal_out, "%s%s = ", isp->external_name, suffix);
	ldf_column = strlen(isp->external_name) + strlen(suffix) + 3;
	start_col = ldf_column;
	if (isp->xor != NULL) {
	    fprintf(pal_out, "(");
	    ++ldf_column;
	    isp_plogic(isp->xor, bp);
	    /* XXX should use $$ in those cases where we can XXX */
	    fprintf(pal_out, ") $ (");
	    ldf_column += 6;
	    isp_plogic(isp->logic, bp);
	    fprintf(pal_out, ");%s", newline);
	} else {
	    isp_plogic(isp->logic, bp);
	    fprintf(pal_out, ";%s", newline);
	}
	if (isp->clock != NULL && !doneclk) {
	    if (isp->flags & PTCLOCK) {
		fprintf(pal_out, "%s%s.PTCLK = ", isp->external_name, suffix);
		ldf_column = start_col + 6;
		if (isp->flags & CLKINV) {
		    fprintf(pal_out, "!(");
		    ldf_column += 2;
		}
		isp_plogic(isp->clock->logic, bp);
		if (isp->flags & CLKINV)
		    fprintf(pal_out, ")");
	    } else {
		fprintf(pal_out, "%s%s.CLK = ",  isp->external_name, suffix);
		ldf_column = start_col + 4;
		if (invert_ctrlfn(isp, CLOCKED))
		    fprintf(pal_out, "!");
		ldf_ctrlfn(isp->clock, CLOCKED);
	    }
	    fprintf(pal_out, ";%s", newline);
	    if (isp->flags & PTRESET) {
		fprintf(pal_out, "%s%s.RE = ", isp->external_name, suffix);
		ldf_column = start_col + 3;
		isp_plogic(isp->reset->logic, bp);
		fprintf(pal_out, ";%s", newline);
	    }
	    doneclk = 1;
	}
    }
    if (bp->ptoe != NULL) {
	fprintf(pal_out, "__OE_%c = ", bp->bank + 'A');
	ldf_column = 9;
	isp_plogic(bp->ptoe->logic, bp);
	fprintf(pal_out, ";%s", newline);
    }
    fprintf(pal_out, "END;%sEND;%s%s", newline, newline, newline);
}

/*
 * Print a logic array to the LDF file.
 */
isp_plogic(ap, bp)
    struct array *ap;
    struct boxinfo *bp;
{
    int i, first, times;
    struct cube *cp;
    struct ispsig *isi;
    int l;

    first = 1;
    for (cp = ap->a_first; cp != NULL; cp = cp->c_next) {
	if (cp->c_logic[n_inputcols] != O)
	    continue;
	if (!first) {
	    fprintf(pal_out, "%s    # ", newline);
	    ldf_column = 6;
	} else
	    first = 0;
	times = 0;
	for (i = 0; i < n_inputcols; ++i) {
	    switch (cp->c_logic[i]) {
	    case O:
	    case Z:
		isi = &isp_sigs[i];
		while (isi->indir != NULL)
		    isi = isi->indir;
		l = strlen(isi->external_name);
		if (times) {
		    if (ldf_column + l + 4 > 75) {
			fprintf(pal_out, "%s     ", newline);
			ldf_column = 5;
		    }
		    fprintf(pal_out, " & ");
		    ldf_column += 3;
		} else
		    times = 1;
		if (cp->c_logic[i] == (isi->flags & USEINVERSE? O: Z)) {
		    fprintf(pal_out, "!");
		    ++ldf_column;
		}
		fprintf(pal_out, "%s", isi->external_name);
		ldf_column += l;
		/* see if we need a .PIN */
		if (bp != NULL && (isi->attr & INTERNAL_SIG)
		    && isi->bank == bp->bank && isi->box == bp->box
		    && isi->clock == NULL && (isi->flags & EXTFEEDBACK) == 0) {
		    fprintf(pal_out, ".PIN");
		    ldf_column += 4;
		}
		break;
	    }
	}
	if (!times)
	    fprintf(pal_out, "1");
    }
    if (first)
	fprintf(pal_out, "0");
}

/*
 * Print the name of a signal used as a control function.
 */
ldf_ctrlfn(cfp, use)
    struct ctrlfn *cfp;
    int use;
{
    char *suffix;

    if (cfp->input_num < 0 || (isp_sigs[cfp->input_num].attr & use) == 0) {
	isp_plogic(cfp->logic, NULL);
	return;
    }
    switch (cfp->usebits) {
    case CLOCKED:
    case INPUT_CLOCK:
	suffix = "_CLK";
	break;
    case ASYNC_RESET:
	suffix = "_RE";
	break;
    case TRISTATE:
	suffix = "_OE";
	break;
    default:
	suffix = "";
    }
    fprintf(pal_out, "%s%s", isp_sigs[cfp->input_num].external_name, suffix);
}

/*
 * Work out whether we need a control input or its inverse.
 */
int
invert_ctrlfn(isp, use)
    struct ispsig *isp;
    int use;
{
    int invbit, ptbit;
    struct ctrlfn *cfp;

    switch (use) {
    case INPUT_CLOCK:
	invbit = INCLKINV;
	ptbit = 0;
	cfp = isp->inclock;
	break;
    case TRISTATE:
	invbit = OEINV;
	ptbit = PTOE;
	cfp = isp->oe;
	break;
    case CLOCKED:
	invbit = CLKINV;
	ptbit = PTCLOCK;
	cfp = isp->clock;
	break;
    default:
	return 0;
    }
    invbit &= isp->flags;
    if (cfp != NULL && cfp->input_num >= 0 && (isp->flags & ptbit) == 0) {
	if (cfp->input_inv)
	    invbit = !invbit;
	if (isp_sigs[cfp->input_num].sig->e_signal->s_attr.a_bits & LOW_TRUE)
	    invbit = !invbit;
    }
    return invbit;
}

/*
 * Yabber on about the device for a bit.
 */
isp_dump()
{
    struct ispsig *isp;
    struct expr *sp;
    struct signal *sig;
    struct ispdecl *eip;
    struct boxinfo *bp;
    int i, use, bit;
    char *sep;

    pal_out = stdout;
    printf("I/O pins:\n");
    for (isp = isp_sigs; isp < &isp_sigs[n_inputcols]; ++isp) {
	if (isp->attr == INTERNAL_SIG)
	    continue;
	sp = isp->sig;
	sig = sp->e_signal;
	printf("%s {%s}: ", signal_name(sp), isp->external_name);
	for (eip = isp->first_decl; eip != NULL; eip = eip->next) {
	    if (eip->num < 0)
		continue;
	    if (eip->attr & CRIT_PATH)
		printf("fast ");
	    switch (eip->attr & ~CRIT_PATH) {
	    case INPUT_BIT:
		printf("input ");
		break;
	    case OUTPUT_BIT:
		printf("output ");
		break;
	    case INPUT_BIT|OUTPUT_BIT:
		printf("I/O ");
		break;
	    case CLOCKED:
		printf("clock ");
		break;
	    case ASYNC_RESET:
		printf("reset ");
		break;
	    default:
		printf("attr=%x ", eip->attr);
	    }
	    if (eip->num > 0)
		printf("(pin %d) ", eip->num);
	}
	if (isp->inclock != NULL)
	    printf("input_clock = %d%s ", isp->inclock - ctrl_fns,
		   (isp->flags & INCLKINV)? " [inverted]": "");
	if (isp->oe != NULL)
	    printf("oe = %d ", isp->oe - ctrl_fns,
		   (isp->flags & OEINV)? " [inverted]": "");
	if (isp->use & INPUT_BIT)
	    printf("used as input ");
	if (isp->use & CLOCKED)
	    printf("used as clock ");
	if (isp->use & INPUT_CLOCK)
	    printf("used as input clock ");
	printf("\n");
    }

    printf("\nControl fns:\n");
    for (i = 0; i < n_ctrlfns; ++i) {
	array_print_eqn(ctrl_fns[i].logic, input_cols, n_inputcols);
	use = ctrl_fns[i].usebits;
	printf(" used as");
	if (use & INPUT_CLOCK)
	    printf(" input clock,");
	if (use & TRISTATE)
	    printf(" output enable,");
	if (use & CLOCKED)
	    printf(" clock,");
	if (use & ASYNC_RESET)
	    printf(" reset,");
	printf("\n");
    }

    printf("\nInternal signals:\n");
    for (isp = isp_sigs; isp < &isp_sigs[n_inputcols]; ++isp) {
	if (isp->attr & INTERNAL_SIG) {
	    printf("%s {%s}: ", signal_name(isp->sig), isp->external_name);
	    if (isp->bank >= 0)
		printf("<%c%c> ", isp->bank + 'A',
		       (isp->box >= 0)? isp->box + '0': '*');
	    if (isp->logic == NULL)
		printf("(no defn)");
	    else {
		if (isp->xor != NULL)
		    printf("1+");
		printf("%d term(s)", count_terms(isp->logic, n_inputcols, O));
		if (isp->flags & INVERTED)
		    printf(" [inverted]");
		printf(", %d inputs", count_bits(isp->sigmap));
	    }
	    if (isp->clock != NULL)
		printf(", %s clock %d%s",
		       (isp->flags & PTCLOCK)? "PT": "global",
		       isp->clock - ctrl_fns,
		       (isp->flags & CLKINV)? " [inverted]": "");
	    if (isp->reset != NULL)
		printf(", %s reset %d%s",
		       (isp->flags & PTRESET)? "PT": "global",
		       isp->reset - ctrl_fns);
	    printf("\n");
	}
    }

    printf("\nBoxes:\n");
    for (bp = box_info; bp < &box_info[isp_dev->n_boxes]; ++bp) {
	if (bp->n_fns > 0 || bp->ptoe != NULL) {
	    printf("%c%d: ", bp->bank + 'A', bp->box);
	    for (i = 0; i < bp->n_fns; ++i) {
		if (i > 0)
		    printf(", ");
		printf("%s", signal_name(bp->funcs[i]->sig));
	    }
	    if (bp->ptoe != NULL) {
		if (i > 0)
		    printf(", ");
		printf("bank OE");
	    }
	    printf("\n   %d inputs", count_bits(bp->sigmap));
	    sep = ": ";
	    bit = -1;
	    while ((bit = next_bit(bp->sigmap, bit)) >= 0) {
		printf("%s%s", sep, signal_name(isp_sigs[bit].sig));
		sep = ", ";
	    }
	    printf("\n");
	}
    }
}

/*
 * Determine whether a given internal signal can be
 * assigned to a given box.
 */
int
can_assign(isp, box)
    struct ispsig *isp;
    struct boxinfo *box;
{
    int i, pn;
    struct ispsig *isi;
    struct bankinfo *kp;
    struct ispdecl *idp;

    if (box->n_fns >= isp_dev->funcs_per_box
	|| (isp->flags & FIXEDBANK) && isp->bank != box->bank
	|| (isp->flags & FIXEDBOX) && isp->box != box->box)
	return 0;
    kp = &bank_info[box->bank];
    if (isp->flags & PTOE) {
	if (kp->ptoe != NULL && kp->ptoe != isp->oe)
	    return 0;
    }
    if ((isp->attr & OUTPUT_BIT) && (isp->flags & FIXEDBANK) == 0
	&& kp->free_ios == 0)
	return 0;
    if ((isp->attr & (OUTPUT_BIT|CRIT_PATH)) == FAST_OUTPUT
	&& box->free_fastos <= 0)
	return 0;
    if (box->n_fns > 0 && isp->clock != NULL) {
	for (i = 0; i < box->n_fns; ++i) {
	    isi = box->funcs[i];
	    if (isi->clock != NULL
		&& (isi->clock != isp->clock
		    || ((isp->flags ^ isi->flags) & (CLKINV|PTRESET)) != 0
		    || (isp->flags & PTRESET) && isi->reset != isp->reset))
		return 0;
	}
    }
    if (isp->attr & OUTPUT_BIT && isp->flags & FIXEDPIN && box->n_fns > 0) {
	pn = 0;
	for (idp = isp->first_decl; idp != NULL; idp = idp->next)
	    if (idp->attr & OUTPUT_BIT) {
		pn = idp->num;
		break;
	    }
	if (pn != 0) {
	    pn = isp_dev->pins[pn - 1].group & 3;
	    for (i = 0; i < box->n_fns; ++i) {
		isi = box->funcs[i];
		if (isi->attr & OUTPUT_BIT && isi->flags & FIXEDPIN) {
		    for (idp = isi->first_decl; idp!= NULL; idp = idp->next)
			if (idp->attr & OUTPUT_BIT) {
			    if (idp->num != 0
				&& (isp_dev->pins[idp->num-1].group & 3) == pn)
				return 0;
			    break;
			}
		}
	    }
	}
    }
    return 1;
}

/*
 * Convert an expression for an internal signal to array form,
 * taking advantage of the hardware XOR if possible.
 */
struct array *
isp_array(exp, ap, xorp, polp)
    struct expr *exp;
    struct attr *ap;
    int *polp;
    struct array **xorp;
{
    struct array *ap1, *ap0;
    struct array *apl0, *apl1, *apr0, *apr1;
    int attr, n1, n0, pol;

    ap0 = NULL;
    ap1 = NULL;
    *xorp = NULL;
    attr = ap->a_bits;
    if ((attr & CRIT_PATH) == 0 && exp->e_op == '^') {
	/* see if we want to use the hardware XOR */
	apl1 = conv_1_expr(exp->e_left, O);
	apr1 = conv_1_expr(exp->e_right, O);
	if (count_terms(apl1, n_inputcols, O) == 1) {
	    exp = exp->e_right;
	    stripx(apl1, n_inputcols);
	    *xorp = apl1;
	    ap1 = apr1;
	} else if (count_terms(apr1, n_inputcols, O) == 1) {
	    exp = exp->e_left;
	    stripx(apr1, n_inputcols);
	    *xorp = apr1;
	    ap1 = apl1;
	} else {
	    apl0 = conv_1_expr(exp->e_left, Z);
	    apr0 = conv_1_expr(exp->e_right, Z);
	    if (count_terms(apl0, n_inputcols, O) == 1) {
		exp = exp->e_right;
		stripx(apl0, n_inputcols);
		*xorp = apl0;
		ap1 = apr0;
		ap0 = apr1;
	    } else if (count_terms(apr0, n_inputcols, O) == 1) {
		exp = exp->e_left;
		stripx(apr0, n_inputcols);
		*xorp = apr0;
		ap1 = apl0;
		ap0 = apl1;
	    } else {
		release(apl0);
		release(apl1);
		release(apr0);
		release(apr1);
	    }
	}
    }

    if (ap1 == NULL)
	ap1 = conv_1_expr(exp, O);
    n1 = count_terms(ap1, n_inputcols, O);
    if (n1 <= 1) {
	pol = O;
	goto done;
    }
    if (ap0 == NULL)
	ap0 = conv_1_expr(exp, Z);
    n0 = count_terms(ap0, n_inputcols, O);
    if (n0 <= 1) {
	pol = Z;
	goto done;
    }

    /*
     * Oh well, we'd better optimize...
     */
    hass(ap1, sigmin(ap1, attr, ap->a_hazard));
    n1 = count_terms(ap1, n_inputcols, O);
    if (n1 <= 3 || *xorp == NULL && n1 <= 4)
	pol = O;
    else {
	hass(ap0, sigmin(ap0, attr, ap->a_hazard));
	n0 = count_terms(ap0, n_inputcols, O);
	pol = (n1 <= n0)? O: Z;
    }

 done:
    if (pol == Z) {
	apl1 = ap1;
	ap1 = ap0;
	ap0 = apl1;
    }
    if (ap0 != NULL)
	release(ap0);
    stripx(ap1, n_inputcols);
    *polp = pol;
    return ap1;
}

struct ctrlfn *
find_ctrlfn(exp, use, flagsp)
    struct expr *exp;
    int use, *flagsp;
{
    int inn, inv, nones, noinv, pol, invbit, attr;
    struct ctrlfn *cfp, *inverse;
    struct array *ap, *ap0;
    struct cube *cp;
    struct ispsig *isi;

    switch (use) {
    case CLOCKED:
	invbit = CLKINV;
	break;
    case TRISTATE:
	invbit = OEINV;
	break;
    case INPUT_CLOCK:
	invbit = INCLKINV;
	break;
    default:
	invbit = 0;
    }

    for (cfp = ctrl_fns; cfp < &ctrl_fns[n_ctrlfns]; ++cfp) {
	if ((exp == cfp->exp || exp == cfp->invexp) && use == cfp->usebits) {
	    if (exp == cfp->invexp)
		*flagsp |= invbit;
	    cfp->usebits |= use;
	    return cfp;
	}
    }

    ap = conv_1_expr(exp, O);
    nones = count_terms(ap, n_inputcols, O);
    pol = O;
    attr = (use==CLOCKED || use==INPUT_CLOCK)? NO_HAZARD: 0;
    if (nones > 1) {
	if (use != ASYNC_RESET) {
	    ap0 = conv_1_expr(exp, Z);
	    noinv = count_terms(ap0, n_inputcols, O);
	    if (noinv == 1) {
		pol = Z;
		release(ap);
		ap = ap0;
	    } else {
		hass(ap, sigmin(ap, attr, NULL));
		hass(ap0, sigmin(ap0, attr, NULL));
		nones = count_terms(ap, n_inputcols, O);
		noinv = count_terms(ap0, n_inputcols, O);
		if (nones < noinv) {
		    release(ap0);
		} else {
		    nones = noinv;
		    pol = Z;
		    release(ap);
		    ap = ap0;
		}
	    }
	} else {
	    hass(ap, sigmin(ap, attr, NULL));
	    nones = count_terms(ap, n_inputcols, O);
	}
    }
    if (pol == Z)
	*flagsp |= invbit;
    sort_array(ap);
    for (cfp = ctrl_fns; cfp < &ctrl_fns[n_ctrlfns]; ++cfp) {
	if (cfp->usebits == use && same_array(ap, cfp->logic)) {
	    release(ap);
	    return cfp;
	}
    }
    ++n_ctrlfns;
    if (pol == O) {
	cfp->exp = exp;
	cfp->invexp = NULL;
    } else {
	cfp->exp = NULL;
	cfp->invexp = exp;
    }
    cfp->logic = ap;
    cfp->usebits = use;
    cfp->nterms = nones;
    if (!single_input(ap, &cfp->input_num, &cfp->input_inv))
	cfp->input_num = -1;
    if (use == TRISTATE || use == ASYNC_RESET || cfp->input_num < 0)
	use = INPUT_BIT;
    note_uses(ap, use);
    return cfp;
}

note_uses(ap, use)
    struct array *ap;
    int use;
{
    struct cube *cp;
    int i;

    for (cp = ap->a_first; cp != NULL; cp = cp->c_next)
	for (i = 0; i < n_inputcols; ++i)
	    if (cp->c_logic[i] != X)
		isp_sigs[i].use |= use;
}

sort_array(ap)
    struct array *ap;
{
    struct cube *cp, **clp, *cnp, *cq;
    int i;

    cp = ap->a_first;
    ap->a_first = NULL;
    for (; cp != NULL; cp = cnp) {
	cnp = cp->c_next;
	if (cp->c_logic[n_inputcols] == O) {
	    for (clp = &ap->a_first; (cq = *clp) != NULL; clp = &cq->c_next) {
		for (i = 0; i < n_inputcols; ++i)
		    if (cp->c_logic[i] != cq->c_logic[i])
			break;
		if (i < n_inputcols && cp->c_logic[i] < cq->c_logic[i])
		    break;
	    }
	    *clp = cp;
	    cp->c_next = cq;
	    if (cq == NULL)
		ap->a_last = cp;
	} else
	    cfree(cp);
    }
}

int
same_array(ap, bp)
    struct array *ap, *bp;
{
    struct cube *cp, *dp;
    int i;

    cp = ap->a_first;
    dp = bp->a_first;
    while (cp != NULL && dp != NULL) {
	for (i = 0; i < n_inputcols; ++i)
	    if (cp->c_logic[i] != dp->c_logic[i])
		return 0;
	cp = cp->c_next;
	dp = dp->c_next;
    }
    return cp == NULL && dp == NULL;
}

int
count_terms(ap, col, v)
    struct array *ap;
    int col, v;
{
    int n;
    struct cube *cp;

    n = 0;
    for (cp = ap->a_first; cp != NULL; cp = cp->c_next)
	if (cp->c_logic[col] == v)
	    ++n;
    return n;
}

stripx(ap, col)
    struct array *ap;
    int col;
{
    struct cube *cp, **clp;

    for (clp = &ap->a_first; (cp = *clp) != NULL; clp = &cp->c_next)
	if (cp->c_logic[col] != O) {
	    *clp = cp->c_next;
	    cfree(cp);
	}
}

/*
 * Test whether the function represented by ap has a single input,
 * possibly inverted.  If not, returns 0.  If so, returns 1, with
 * the index of the input signal in *innp, with *invp indicating
 * if the signal is inverted.
 */
int
single_input(ap, innp, invp)
    struct array *ap;
    int *innp, *invp;
{
    int i, inn;
    struct cube *cp;

    inn = -1;
    for (cp = ap->a_first; cp != NULL; cp = cp->c_next)
	if (cp->c_logic[n_inputcols] == O) {
	    if (inn >= 0)
		return 0;
	    for (i = 0; i < ap->a_noin; ++i)
		if (cp->c_logic[i] != X) {
		    if (inn >= 0)
			return 0;
		    inn = i;
		    *invp = cp->c_logic[i] == Z;
		}
	}
    if (inn < 0)
	return 0;
    *innp = inn;
    return 1;
}

/*
 * Make a string name for a signal.
 */
char *
signal_name(sp)
    struct expr *sp;
{
    struct signal *sig;
    static char buf[512];

    if (sp->e_op == SIGNAL) {
	sig = sp->e_signal;
	if (sig->s_ix.i_nel == 1 && sig->s_ix.i_minix == 0)
	    return sig->s_ident->i_name;
	sprintf(buf, "%s[%d]", sig->s_ident->i_name, sp->e_index);
    } else
	sprintf(buf, "?BUG: op=%d", sp->e_op);
    return buf;
}

int
count_bits(bm)
    unsigned long *bm;
{
    int i, n;
    unsigned long v;

    n = 0;
    for (i = bitmap_width; i > 0; --i) {
	v = *bm++;
	while (v != 0) {
	    ++n;
	    v &= v - 1;
	}
    }
    return n;
}

add_bits(b1, b2)
    unsigned long *b1, *b2;
{
    int i;

    for (i = 0; i < bitmap_width; ++i)
	b1[i] |= b2[i];
}

int
remove_bits(b1, b2)
    unsigned long *b1, *b2;
{
    int i, ret;

    ret = 0;
    for (i = 0; i < bitmap_width; ++i)
	if ((b1[i] &= ~b2[i]) != 0)
	    ret = 1;
    return ret;
}

/*
 * Count the number of bits in (b1 + b2) - b3.
 */
int
count_sum_bits(b1, b2, b3)
    unsigned long *b1, *b2, *b3;
{
    bitmap sum;
    int i;

    for (i = 0; i < bitmap_width; ++i)
	sum[i] = (b1[i] | b2[i]) & ~b3[i];
    return count_bits(sum);
}

/*
 * Set bd to the intersection of b1 and b2.
 * Return 0 if the intersection is empty, 1 otherwise.
 */
int
intersect_maps(bd, b1, b2)
    unsigned long *bd, *b1, *b2;
{
    int i, res;

    res = 0;
    for (i = 0; i < bitmap_width; ++i)
	if ((bd[i] = b1[i] & b2[i]) != 0)
	    res = 1;
    return res;
}

/*
 * Return the index of the next bit set in `map' after `bit',
 * or -1 if there are no more.  (This could be optimized.)
 */
int
next_bit(map, bit)
    unsigned long *map;
    int bit;
{
    while (++bit < n_inputcols)
	if (ISSET(map, bit))
	    return bit;
    return -1;
}
